import matplotlib.pyplot as plt
from typing import List, Dict, Union, Tuple, Callable
import os
import math
import numpy as np
import pandas as pd
from .common_labels import common_labels
from scipy.signal import savgol_filter


def get_alias(label: str):
    for name, labels in common_labels.items():
        if label in labels:
            return name
    return label


def filter_signal(y: List[float], window: int = 17, polyorder: int = 3) -> List[float]:
    window = max(7, min(window, len(y)))
    if window % 2 == 0:
        window -= 1
    if len(y) < window:
        return y
    return savgol_filter(y, window, polyorder)


def get_figsize(n: int, graphs_per_row: int) -> Tuple[int, int]:
    return min(n, graphs_per_row), math.ceil(n/graphs_per_row)


def _plot_history(histories: pd.DataFrame, style: str = "-", interpolate: bool = False, side: float = 5, graphs_per_row: int = 4, customization_callback: Callable = None, path: str = None):
    """Plot given training histories.
        histories:pd.DataFrame, the histories to plot.
        style:str="-", the style to use when plotting the graphs.
        interpolate:bool=False, whetever to reduce the graphs noise.
        side:int=5, the side of every sub-graph.
        graphs_per_row:int=4, number of graphs per row.
        customization_callback:Callable=None, callback for customising axis.
        path:str=None, where to save the graphs, by defalut nowhere.
    """
    x_label = "Epochs" if histories[0].index.name is None else histories[0].index.name
    metrics = [m for m in histories[0] if not m.startswith("val_")]
    n = len(metrics)
    w, h = get_figsize(n, graphs_per_row)
    _, axes = plt.subplots(h, w, figsize=(side*w, side*h))
    flat_axes = np.array(axes).flatten()

    average_history = pd.concat(histories)
    average_history = average_history.groupby(average_history.index).mean()

    for i, history in enumerate([average_history] + histories):
        for metric, axis in zip(metrics, flat_axes):
            for name, kind in zip((metric, f"val_{metric}"), ("Train", "Test")):
                if name in history:
                    col = history[name]
                    if i == 0:
                        axis.plot(
                            col.index.values,
                            filter_signal(
                                col.values) if interpolate else col.values,
                            style,
                            label='{kind}: {val:0.4f}'.format(
                                kind=kind, val=col.iloc[-1]),
                            zorder=10000
                        )
                    else:
                        axis.plot(
                            col.index.values,
                            filter_signal(col.values) if interpolate else col.values,
                            style,
                            alpha=0.3
                        )

    for metric, axis in zip(metrics, flat_axes):
        alias = get_alias(metric)
        axis.set_xlabel(x_label)
        axis.set_ylabel(alias)
        axis.set_title(alias)
        axis.grid(True)
        axis.legend()
        if metric in ("auprc", "auroc", "acc", "accuracy"):
            axis.set_ylim(-0.05, 1.05)
        if history.shape[0] <= 4:
            axis.set_xticks(range(history.shape[0]))
        if customization_callback is not None:
            customization_callback(axis)

    for axis in flat_axes[len(metrics):]:
        axis.axis("off")

    plt.tight_layout()
    if path is not None:
        plt.savefig(path)


def _get_columns(history: pd.DataFrame) -> List[str]:
    return [[c] if f"val_{c}" not in history else [c,  f"val_{c}"] for c in history if not c.startswith("val_")]


def filter_column(histories: List[str], columns: List[str]) -> List[pd.DataFrame]:
    return [history[columns] for history in histories]


def plot_history(histories: Union[Dict[str, List[float]], pd.DataFrame, List[pd.DataFrame]], style: str = "-", interpolate: bool = False, side: float = 5, graphs_per_row: int = 4, customization_callback: Callable = None, path: str = None, single_graphs: bool = False):
    """Plot given training histories.
        histories:Union[Dict[str, List[float]], pd.DataFrame, List[pd.DataFrame]], the histories to plot.
        style:str="-", the style to use when plotting the graphs.
        interpolate:bool=False, whetever to reduce the graphs noise.
        side:int=5, the side of every sub-graph.
        graphs_per_row:int=4, number of graphs per row.
        customization_callback:Callable=None, callback for customising axis.
        path:str=None, where to save the graphs, by defalut nowhere.
        single_graphs:bool=False, whetever to create the graphs one by one.
    """
    if not isinstance(histories, list):
        histories = [histories]
    if path is not None:
        dirname = os.path.dirname(path)
        if dirname:
            os.makedirs(os.path.dirname(path), exist_ok=True)
    histories = [
        pd.DataFrame(history) if not isinstance(history, pd.DataFrame) else history for history in histories
    ]
    if single_graphs:
        for columns in _get_columns(histories[0]):
            _plot_history(
                filter_column(histories, columns),
                style,
                interpolate,
                side,
                graphs_per_row,
                customization_callback,
                "{path}/{c}.png".format(path=path, c=columns[0])
            )
    else:
        _plot_history(histories, style, interpolate, side, graphs_per_row,
                      customization_callback, path)



from typing import List, Dict

def chain_histories(new_history:Dict[str, List[float]], old_history:Dict[str, List[float]]):
    if old_history is None:
        return new_history
    return {
        key: old_history[key] + new_history[key] for key in old_history
    }


from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open as copen
from os import path
import os
import re

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with copen(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


def read(*parts):
    with copen(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


__version__ = find_version("plot_keras_history", "__version__.py")

test_deps = ['pytest', 'pytest-cov', 'coveralls', 'validate_version_code', 'codacy-coverage']

extras = {
    'test': test_deps,
}


setup(
    name='plot_keras_history',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=__version__,

    description='A simple python package to print a keras NN training history.',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/LucaCappelletti94/plot_keras_history',

    # Author details
    author='Luca Cappelletti',
    author_email='cappelletti.luca94@gmail.com',

    # Choose your license
    license='MIT',

    include_package_data=True,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Information Technology',
        'Topic :: Scientific/Engineering :: Information Analysis',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3'
    ],
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "matplotlib",
        "pandas",
        "scipy"
    ],
    tests_require=test_deps,
    extras_require=extras,
)